;;; ACT-R Transfer (Actransfer)
;;;
;;; Copyright 2011-2014 Niels Taatgen
;;; 
;;; Version 0.55
;;; Fixed several bugs in spreading activation
;;;
;;; Version 0.54
;;;  Fixed LHS RTid
;;;
;;; Version 0.53
;;; Automatic reordering of conditions (i.e., WM1=V1 becomes V1=WM1)
;;;
;;; version 0.52
;;; Set action time of harvest retrieval to 0
;;; Give an option to only compile complete task-specific rules *task-specific-compile* can be condition-only or complete
;;;
;;; version 0.51
;;; Bug fix in compiling +update-perception>
;;; Version 0.5 Cleaned up the code that initiates all the module actions
;;;           Some syntax improvements: operator instread of ins and add-task
;;; 
;;; Version 0.45 Bug-fix

#-:act-r-6.0 (error "Please load ACT-R 6.0 before loading ACTransfer")
#+:actransfer-0.55 (error "ACTransfer should be loaded once")
#-:actransfer-0.55 (pushnew :actransfer-0.55 *features*)

(defconstant *actransfer-version* "0.55")
(defconstant *actransfer-date* "15 May 2014")


(defvar *exclude-from-ul*)

(setf *exclude-from-ul* '(retrieve-instruction store-retrieved-instruction general-condition-does-not-match general-action-does-not-apply store-retrieved-instruction start-action-sequence action-goal-finish update-perception-on-wait))

(defun linear-update-utility (module production reward)
  (unless (member production *exclude-from-ul*)  ;;; Added
  (let ((old (production-u production))
        (alpha (utility-alpha module)))
    (setf (production-u production) (+ old (* alpha (- reward old)))))))

;;; Standard production compilation can misbehave in this model, because it relies on declarative
;;; finst to track instructions we have tried. So we only want to compile an instruction if all
;;; conditions are tested.

;;; Hack compile-productions to prevent compilation of retrieve-instruction unless it is with a production that clears operator
;;; Also, we don't want retrieve-instruction as a second rule in a compilation
;;;


(defvar *task-specific-compile*)
(setf *task-specific-compile* 'condition-only) ;;; can be condition-only or complete

(defun condition-nil-action (production)
  (let* ((prod (produce-standard-representation production))
         (action (second prod))
         (goal-action (second (assoc '=goal> action))))
    (cond ((eq *task-specific-compile* 'condition-only)
           (or (member '(action wait-all-done) goal-action :test #'equal)
               (member '(action nil) goal-action :test #'equal)
               (member '(condition nil) goal-action :test #'equal)
               (member '(condition act) goal-action :test #'equal)))
          ((eq *task-specific-compile* 'complete) (member '(action nil) goal-action :test #'equal))
          (t (model-warning "Invalid value for *task-specific-compile*")))
))
         
(defun check-condition-action-bridge (p1 p2)
  (let* ((p1-action (second (assoc '=goal> (second (produce-standard-representation (get-production (first p1)))))))
         (p2-condition (second (assoc '=goal> (first (produce-standard-representation p2)))))
         (p2-action (second (assoc '=goal> (second (produce-standard-representation p2))))))
    (and ; (not (assoc 'task p2-condition))
         (member '(= condition act) p1-action :test #'equal)
         (not (or (member '(action pending) p2-action :test #'equal)
                  (member '(action nil) p2-action :test #'equal))))))

(defun check-action-perception (p1 p2)
  (let* ((p1-action (second (assoc '+action> (second (produce-standard-representation p1)))))
         (p2-condition (second (assoc '=perception> (first (produce-standard-representation p2))))))
    (and p1-action p2-condition)))

(defun check-fact-retrieval (p1 p2)
   (let* ((p1-action (second (assoc '+retrieval> (second (produce-standard-representation p1)))))
          (p2-condition (second (assoc '=retrieval> (first (produce-standard-representation p2))))))
     (and p2-condition (member '(isa fact) p2-condition :test #'equal))))

(defun check-double-update-perception (p1 p2)
   "Yields true if both productions do a +perception> isa update-perception, otherwise nil"
    (let* ((p1-action (second (assoc '+perception> (second (produce-standard-representation p1)))))
           (p2-action (second (assoc '+perception> (first (produce-standard-representation p2)))))
           (p2-condition (assoc '=perception> (first (produce-standard-representation p2))))
           (check (and (null p2-condition)(equal p1-action '(isa update-perception))(equal p2-action '(isa update-perception)))))
      (when (not check) (model-output "  Buffer perception prevents composition of these productions"))
      check))
      

(defun composeable-productions-p (module p1 p2)
  (multiple-value-bind (val exists)
      (gethash (vector (production-name p1) (production-name p2)) (compilation-module-composeable-table module))
    (if (and exists (listp val) (eq (first val) 'perception)) ;t   ;;;; Too general, fix!!! This version ignores all perception collapses
      (check-double-update-perception p1 p2)
    (if exists
        (cond
         ((eq val t)
          t)
         ((listp val)
          (when (compilation-module-trace module)
            (if (first val)
                (model-output "  Buffer ~S prevents composition of these productions~@[~%   because ~a~]" (first val) (second val))
              (model-output "  ~a prevents composition of these productions" (second val)))))
         (t
          (print-warning "Invalid value in composeable table -- contact Dan.")))
      (let ((val (determine-composable module p1 p2)))
        (setf (gethash (vector (production-name p1) (production-name p2)) (compilation-module-composeable-table module)) val)
        (if (eq val t) t nil))))))

(defvar *task-specific-compile*)
(setf *task-specific-compile* 'condition-only) ;;; can be condition-only or complete

(defun compile-productions (production)
  (let ((module (get-module production-compilation))
        (p-name (production-name production)))
    
    (when (compilation-module-epl module)
      
      (when (compilation-module-trace module)
        (model-output "Production Compilation process started for ~s" p-name))
      
      (cond  ((equal p-name 'retrieve-instruction)
             (when (compilation-module-trace module)
               (model-output "  Production ~s can never compile as second production" p-name))
             (handle-check-valid-1st-p module p-name production))  ;;; NT: added this condition
 
             ((not (valid-compilation-production p-name production))
             (when (compilation-module-trace module)
               (model-output "  Production ~s is not valid for compilation" p-name))
             (setf (compilation-module-previous module) nil))
            
            ((null (compilation-module-previous module))
             (when (compilation-module-trace module)
               (model-output "  No previous production to compose with."))
             (handle-check-valid-1st-p module p-name production))

            ((check-action-perception (get-production (first (compilation-module-previous module))) production)
              (when (compilation-module-trace module)
               (model-output "  No compilation because +action> cannot be followed by =perception>."))
             (handle-check-valid-1st-p module p-name production))  ;;; NT: added this condition
            
            ((check-fact-retrieval (get-production (first (compilation-module-previous module))) production)
             (when (compilation-module-trace module)
               (model-output "  No compilation because we do not yet compile over retrievals of type fact."))
             (handle-check-valid-1st-p module p-name production))  ;;; NT: added this condition

            ((and (equal (car (compilation-module-previous module)) 'retrieve-instruction)  ;;; Retrieve-instruction is the first
                  ;;; Is there a =goal> condition nil action in the production?
                  (not (condition-nil-action production)))
              (when (compilation-module-trace module)
               (model-output "  No compilation because retrieve-instruction does not combine with productions that don't finalize a condition."))
             (handle-check-valid-1st-p module p-name production))  ;;; NT: added this condition
            
            ((check-condition-action-bridge (compilation-module-previous module) production)
              (when (compilation-module-trace module)
               (model-output "  No compilation because of improper condition-action bridge."))
             (handle-check-valid-1st-p module p-name production))  ;;; NT: added this condition
 

            ((> (- (mp-time) (compilation-module-previous-time module))
                (compilation-module-tt module))
             (when (compilation-module-trace module)
               (model-output "  Cannot compile ~s and ~s because the time between them exceeds the threshold time."
                             (car (compilation-module-previous module))
                             p-name))
             (handle-check-valid-1st-p module p-name production))
            ((null (composeable-productions-p module (get-production (car (compilation-module-previous module))) production))
             (when (compilation-module-trace module)
               (model-output "  Production ~s and ~s cannot be composed." (car (compilation-module-previous module)) p-name))
             (handle-check-valid-1st-p module p-name production))
            (t
             (when (compilation-module-trace module)
               (model-output "  Production ~s and ~s are being composed." (car (compilation-module-previous module)) p-name))
             (compose-productions module production)
             (handle-check-valid-1st-p module p-name production))))))

;;; Change compose-productions to make it drop queries that are no longer necessary.
;;; For example, a ?retrieval> state free is dropped if the action does not contains a +retrieval>
;;; same for ?action> and ?imaginal>
;;; But first test in on retrieval

(defun compose-productions (module p2)
  ;(print-warning "Composition process not yet operational")
  
  (let* ((p1 (get-production (first (compilation-module-previous module))))
         (p1-s (produce-standard-representation p1))
         (p2-s (produce-standard-representation p2))
         (p1-variables (production-variables p1))
         (p2-variables (production-variables p2))
         (p1-basic-vars (remove-buffers p1-variables))
         (p2-basic-vars (remove-buffers p2-variables))
         (p2-name (production-name p2))
         (new-bindings nil))
    
    (setf (production-compilation-instan p2-name) (production-bindings p2))
    
    (awhen (intersection p1-basic-vars p2-basic-vars)
           (multiple-value-setq (p2-s new-bindings) 
             (rename-variables p2-s it (union p1-basic-vars p2-basic-vars)))
           
           (setf (production-compilation-instan p2-name) 
             (replace-variables (production-compilation-instan p2-name) new-bindings))
           
           (setf (production-drop-out-buffers-map p2-name)
             (remove-if-not (lambda (x) (find x *buffer-var-names*))
                            new-bindings  :key #'car)))
    
    
    ;; To allow for proper mapping of variables and values between
    ;; p1 and p2 it may be necessary to instantiate slot name variables
    ;; in dynamic productions.
    ;;
    ;; This should only be necessary if a buffer allows for modifications on
    ;; the RHS of p1 which could be carried over to the conditions of p2 for now, but
    ;; there may be other conditions that are discovered in the future.
    ;;
    ;; The compilation type needs to indicate if that is possible and may also
    ;; specify a constraint on when to allow it to happen (which presumably is
    ;; just for efficency because conditional application of this for other 
    ;; reasons could lead to unusual results in the newly composed production).
    ;;
    
    
    ;(pprint p1-s)
    ;(pprint p2-s)
    
    (dolist (b (production-buffer-indices p1))
      (let* ((comp-type (get-compilation-type-struct (car b) module))
             (instantiate (comp-buffer-type-pre-instantiate comp-type)))
        
        (when (and instantiate (or (eq instantiate t) (funcall instantiate b p2)))
          
          (let* ((b= (intern (concatenate 'string "=" (string (car b)) ">")))
                 (b+ (intern (concatenate 'string "+" (string (car b)) ">")))
                 (c2 (copy-tree (find b= (first p2-s) :key #'car)))
                 (a1= (copy-tree (find b= (second p1-s) :key #'car)))
                 (a1* (copy-tree (aif (find b+ (second p1-s) :key #'car)
                                      (if (eq 'isa (caaadr it))
                                          it nil)
                                      nil)))
                 (mod-vars (append (mapcan (lambda (x)
                                             (when (chunk-spec-variable-p (car x))
                                               (list (car x))))
                                     (cadr a1=))
                                   (mapcan (lambda (x)
                                             (when (chunk-spec-variable-p (car x))
                                               (list (car x))))
                                     (cadr a1*))))
                 (cond-vars (mapcan (lambda (x)
                                      (when (chunk-spec-variable-p (second x))
                                        (list (second x))))
                              (cdadr c2))))
            
            ;
            ;(pprint b=)
            ;(pprint c2)
            ;(pprint (cdadr c2))
            ;(pprint a1=)
            
            ;(pprint mod-vars)
            ;(pprint cond-vars)
            
            ;(pprint (second (compilation-module-previous 
            ;                 (get-module production-compilation))))
            ;(pprint p2-instans)
            
            
            (cond ((and mod-vars cond-vars)
                   (let* ((p1-instantiations (second (compilation-module-previous 
                                                      (get-module production-compilation))))
                          
                          
                          (s1 (remove-if-not (lambda (x) (find x mod-vars)) p1-instantiations :key #'car))
                          (s2 (remove-if-not (lambda (x) (find x cond-vars)) (production-compilation-instan p2-name) :key #'car)))
                     
                     (dolist (mv s1)
                       (dolist (cv (remove-if-not (lambda (x) (eq (cdr mv) x)) s2 :key #'cdr))
                         
                         (setf p1-s (replace-variables-special p1-s (list mv)))
                         (setf p2-s (replace-variables-special p2-s (list cv)))))))
                  ((and mod-vars (cdadr c2))  ;; variables in the modification and any slots tested in c2
                   (let* ((p1-instantiations (second (compilation-module-previous 
                                                      (get-module production-compilation))))
                          
                          
                          (s1 (remove-if-not (lambda (x) (find x mod-vars)) p1-instantiations :key #'car)))
                     
                     (dolist (mv s1)
                       (setf p1-s (replace-variables-special p1-s (list mv))))))
                  (cond-vars
                   (let ((s2 (remove-if-not (lambda (x) (find x cond-vars)) (production-compilation-instan p2-name) :key #'car)))
                     (dolist (cv s2)
                       (setf p2-s (replace-variables-special p2-s (list cv)))))))))))
    
    ;(pprint p1-s)
    ;(pprint p2-s)
    
    (let* ((mappings nil)
           (ppm (compilation-module-ppm module))
           (bindings (when ppm 
                     (append (second (compilation-module-previous module))
                             (production-compilation-instan (production-name p2))))))
      
      (dolist (buffer (union (mapcar #'car (production-buffer-indices p1))
                             (mapcar #'car (production-buffer-indices p2))))
        ;(pprint buffer)
        (let* ((comp-type (get-compilation-type-struct buffer module))
               (map-fn (comp-buffer-type-map comp-type)))
          
          (when map-fn
             (setf mappings (append (funcall map-fn module p1 p1-s p2 p2-s buffer) mappings)))
        ;(pprint mappings)
        ))
      
      ;(pprint mappings)
      
      ;;; Current mechanism for !safe-bind! on the RHS of P1 is to
      ;;; add its binding as a constant value and drop the bind.
            
      ;;; Here we add the binding to the mappings
      
      (dolist (x (second p1-s))
        (when (eq (car x) '!safe-bind!)
          ;(pprint x)
          (push-last (assoc (car (second x)) 
                            (second (compilation-module-previous module)))
                     mappings)
          ;(pprint 'variables)
          (dolist (y (find-variables (cdr (second x))))
            (push-last (assoc y (second (compilation-module-previous module)))
                       mappings))
          ;(pprint 'not-variables)
          ))
      
      
      (setf mappings (remove-duplicates mappings))
      (setf mappings (sort mappings #'< :key #'(lambda (x)
                                                 (if (constant-value-p (cdr x))
                                                     (if (constant-value-p (car x))
                                                         1 
                                                       2)
                                                   0))))
      
      
      ;(pprint mappings)
      ;(pprint p1-s)
      ;(pprint p2-s)
      
      (loop
        (let* ((mapping (pop mappings))
               (const (and ppm (constant-value-p (cdr mapping)))))
          (when (null mapping) (return))
          
          
          (when const
            (awhen (find (car mapping) bindings :key 'car)
                   ;(format t "Changing mapping from :~s to: ~S~%" mapping it)
                   (setf mapping it)))
          
          ;; this doesn't replace constants which is a good thing...
          
          (setf p1-s (replace-variables-special p1-s (list mapping)))
          (setf p2-s (replace-variables-special p2-s (list mapping)))
          
         ; (pprint const)
          
          
          (dolist (x mappings)
            ;  Don't want to do any constant to constant substitutions...
            ;  at least not with current plans
            ;(if const
             ;   (cond ((and (eq (car x) (car mapping)) (eq (cdr x) (car mapping))) ;; crazy case, but better be safe
             ;          (pprint 1)
             ;          (replace mappings (list (cons (cdr mapping) (cdr mapping))) :start1 (position x mappings :test #'equal)))
             ;         ((eq (car x) (car mapping))
             ;          (pprint 2)
             ;          (replace mappings (list (cons (cdr mapping) (cdr x))) :start1 (position x mappings :test #'equal)))
             ;         ((eq (cdr x) (car mapping))
             ;          (pprint 3)
             ;          (replace mappings (list (cons (car x) (cdr mapping))) :start1 (position x mappings :test #'equal)))
             ;         )
              (when (eq (car x) (car mapping))
                (replace mappings (list (cons (cdr mapping) (cdr x))) :start1 (position x mappings :test #'equal)))
            ;  )
            )
          
         ; (pprint mapping)
         ; (pprint p1-s)
          ;(pprint p2-s)
          ))
      
      ;(pprint p1-s)
      ;(pprint p2-s)
      
      (let ((p3-s (list nil nil)))
        (dolist (buffer (union (mapcar #'car (production-buffer-indices p1))
                               (mapcar #'car (production-buffer-indices p2))))
          
          ;(pprint buffer)
          
          (let* ((comp-type (get-compilation-type-struct buffer module))
                 (compose-fn (comp-buffer-type-compose comp-type)))
          
          (when compose-fn
             (let ((vals (funcall compose-fn p1 p1-s p2 p2-s buffer)))
                 ;(pprint it)
                 ;(pprint vals)
                 (setf (first p3-s) (append (first p3-s) (first vals)))
                 (setf (second p3-s) (append (second p3-s) (second vals)))))))
        
        
        ;; Copy over anything that isn't a buffer reference
        
        ;; Making sure that the p1 stuff comes before the
        ;; p2 stuff in the new production (since it's being
        ;; pushed it must go backwards)
        
        
        (dolist (x (first p2-s))
          (when (find (car x) '(!safe-eval!))
            (push x (first p3-s))))
        
        ;;; need to make sure binds come before evals from the p2 stuff
        ;;; so just do it as a quick and dirty two pass now
        ;;; original order might be nice, but this should handle things
        ;;; for now since that gets munged in the standard rep...
        
        (dolist (x (second p2-s))
          (when (find (car x) '(!safe-bind!))
            (push x (second p3-s))))
        
        (dolist (x (second p2-s))
          (when (find (car x) '(!safe-eval! !output!))
            (push x (second p3-s))))
        
        
        (dolist (x (first p1-s))
          (when (find (car x) '(!safe-eval!))
            (unless (find x (first p3-s) :test #'equal)
              (push x (first p3-s)))))
        
        ;;; Here's where we remove the bind.
        
        (dolist (x (second p1-s))
          (when (find (car x) '(#|!safe-bind!|# !safe-eval! !output! !stop!))
            (unless (find x (second p3-s) :test #'equal)
              (push x (second p3-s)))))
        
        
        ;(pprint p3-s)
        
        ;; Double check that everything gets bound 
        ;; The assumption being that it had to come from the second
        ;; production to get lost (through a retrieval removal or 
        ;; colapsing of conditions).  So, just replace it with
        ;; the instantiated value from p2 if there are any.
        
        (let ((lhs-vars (find-variables (mapcar #'cdr (first p3-s))))
              (rhs-vars (find-variables (mapcar #'cdr (second p3-s))))
              (rhs-bindings (mapcar #'caadr (remove '!safe-bind! (second p3-s) :test-not #'eql :key #'car))))
          
          ; (pprint lhs-vars)
          ; (pprint rhs-vars)
          ; (pprint rhs-bindings)
          
          (dolist (v rhs-vars)
            (unless (or (find v *buffer-var-names*) ;; don't want to bind the buffer refs...
                        (find v lhs-vars)
                        (find v rhs-bindings)
                        )
              
              ;(pprint v)
              (setf (second p3-s) (replace-variables (second p3-s) (list (assoc v (production-compilation-instan p2-name))))))))
        
        
 ;       (pprint p3-s)

        ;;;; NT: Added for Actransfer: remove ?retrieval> conditions if there is no +retrieval> action
        ;;;; Possible glitch is if the test is for retrieval errors. But we should compile over those anyway.
        ;;;; Also don't remove the ?retrieval if there is a =rrequest> action, because that means we're setting up a retrieval.
        ;;; update: only do this for "complete productions", i.e., those that test all conditions and do all actions
        ;;; How to know what those are?
        ;;; - they include a check-done-check-modules --> set retrieval-request to nil

        (let* ((p3-c (first p3-s))
              (p3-a (second p3-s))
              (goal-test (second (assoc '=goal> p3-a))))
          (when (and (assoc '?retrieval> p3-c)(assoc '=retrieval> p3-c))
            (setf p3-c (remove-if #'(lambda (x) (eq (first x) '?retrieval>)) p3-c)))
          (when (member '(retrieval-request nil) goal-test :test #'equal)
            (when (and (assoc '?retrieval> p3-c) (not (assoc '+retrieval> p3-a))(not (assoc '=rrequest> p3-a))(not (assoc '=rharvest> p3-c)))
              (setf p3-c (remove-if #'(lambda (x) (eq (first x) '?retrieval>)) p3-c)))
            (when (and (assoc '?imaginal> p3-c) (not (assoc '+imaginal> p3-a)))
              (setf p3-c (remove-if #'(lambda (x) (eq (first x) '?imaginal>)) p3-c)))
            (when (and (assoc '?action> p3-c) (not (assoc '+action> p3-a)))
              (setf p3-c (remove-if #'(lambda (x) (eq (first x) '?action>)) p3-c))))            

            (setf p3-s (list p3-c p3-a)))

 ;       (pprint p3-s)


        ;;; NT end of addition


        (let ((post-conditions nil)
              (new-p nil)
              (dynamic nil))
          
          
          (push (new-production-name) new-p)
          
          (push (format nil "~A & ~A~{~@[ - ~a~]~}"
                  (production-name p1) (production-name p2)
                  (mapcan #'(lambda (x)
                              (let ((comp-type (get-compilation-type-struct x module)))
                                (when (and (comp-buffer-type-drop-out comp-type)
                                           (not (zerop (logand (cdr (assoc x (production-buffer-indices p1))) 4))))  ;; require the request bit be set in the first production's usage
                                  (list (aif (chunk-copied-from-fct (cdr (assoc (read-from-string (format nil "=~a" x)) (production-bindings p2))))
                                             it
                                             (cdr (assoc (read-from-string (format nil "=~a" x)) (production-bindings p2))))))))
                    (production-lhs-buffers p2)))
                new-p)
          
          (dolist (condition (sort-conditions (first p3-s)))
            
            ;(pprint condition)
            
            (if (find (car condition) '(!safe-bind! !safe-eval! !output! !stop!))
                (progn
                  (push (car condition) post-conditions)
                  (push (first (second condition)) post-conditions))
              
              (progn
                (push (car condition) new-p)
                (dolist (test (second condition))
                  (when (and (= (length test) 3)
                             (chunk-spec-variable-p (second test)))
                    (setf dynamic t))
                  
                  (dolist (x (if (and (= (length test) 3)
                                      (eq (car test) '=))
                                 (cdr test)
                               test))
                    (push x new-p)))))
            )
          
          
          (when post-conditions
            (setf new-p (append post-conditions new-p)))
          
          
          (push '==> new-p)
          
          
          (dolist (action (sort-actions (second p3-s)))
            
            ;(pprint action)
            
            (push (car action) new-p)
            
            
            
            (cond ((find (car action) '(!safe-eval! !output!))
                   (push (first (second action)) new-p))
                  ((eq (car action) '!safe-bind!)
                   (push (first (second action)) new-p)
                   (push (second (second action)) new-p))
                  ((null (second action))
                   )
                  ((and (= (length (second action)) 1)
                        (not (listp (first (second action)))))
                   (push (first (second action)) new-p))
                  (t
                   (dolist (test (second action))
                     
                     (when (or (and (= (length test) 3)
                                    (chunk-spec-variable-p (second test)))
                               (and (= (length test) 2)
                                    (chunk-spec-variable-p (first test))))
                       (setf dynamic t))
                     
                     (dolist (x (if (and (= (length test) 3)
                                         (eq (car test) '=))
                                    (cdr test)
                                  test))
                       (push x new-p))))))
          
          ;(pprint (reverse new-p))
          
          (let ((procedural (get-module procedural)))
            (setf (procedural-delay-tree procedural) t)
            
            (let* ((new-prod (if (and (or (production-dynamic p1) (production-dynamic p2)) dynamic) ;; double-check
                                 (with-unchecked-p* (p*-fct  (reverse new-p)))
                               (p-fct (reverse new-p))))
                   (p3 (get-production new-prod)))
              (setf (procedural-delay-tree procedural) nil)
              
              (when p3
                (update-params-for-compiled-production p3 new-prod p1 p2 module procedural)))))))))


;;; Automatically copy the name of a new fact chunk created in the imaginal to its id slot
;;;



(defun fill-in-id-slot ()
  (let ((chunk (act-r-buffer-chunk (buffer-instance 'imaginal))))
    (mod-buffer-chunk 'imaginal (list 'id chunk))))
      

(defun imaginal-request-2 (instance buffer-name chunk-spec)
  (if (imaginal-module-busy instance)
    (model-warning "Imaginal request made to the ~S buffer while the imaginal module was busy. New request ignored." buffer-name)
  (progn
    (setf (imaginal-module-busy instance) t)
    (setf (imaginal-module-error instance) nil)
    
    (case buffer-name
      (imaginal
       (let ((delay (if (imaginal-module-randomize instance)
                        (randomize-time (imaginal-module-delay instance))
                      (imaginal-module-delay instance))))
         
         (schedule-event-relative delay
                                  'set-imaginal-free
                                  :module 'imaginal
                                  :output nil
                                  :priority -1020)
         
         (schedule-event-relative delay
                                  'goal-style-request
                                  :params (list 'imaginal chunk-spec)
                                  :destination 'imaginal
                                  :module 'imaginal
                                  :output nil)
;; Added

      (when (eq (chunk-spec-chunk-type chunk-spec) 'fact)
        (schedule-event-relative delay 
                                 'fill-in-id-slot
                                 :module 'imaginal
                                 :output nil
                                 :priority -1010))))
;; End of addition

      (imaginal-action
       (case (chunk-spec-chunk-type chunk-spec)
         (generic-action
          (let ((action-spec (chunk-spec-slot-spec chunk-spec 'action)))
            
            (cond ((null action-spec)
                   (print-warning "An imaginal-action generic-action request requires an action.")
                   (set-imaginal-free))
                  ((> (length action-spec) 1)
                   (print-warning "An imaginal-action generic-action request requires a single action.")
                   (set-imaginal-free))
                  ((not (eq '= (caar action-spec)))
                   (print-warning "An imaginal-action generic-action request requires the = specifier for the action.")
                   (set-imaginal-free))
                  ((not (or (functionp (third (car action-spec))) (fboundp (third (car action-spec)))))
                   (print-warning "An imaginal-action generic-action request requires the action to name a valid function.")
                   (set-imaginal-free))
                  (t
                   (funcall (third (car action-spec)))))))
         (simple-action
          (let ((action-spec (chunk-spec-slot-spec chunk-spec 'action)))
            
            (cond ((null action-spec)
                   (print-warning "An imaginal-action simple-action request requires an action.")
                   (set-imaginal-free))
                  ((> (length action-spec) 1)
                   (print-warning "An imaginal-action simple-action request requires a single action.")
                   (set-imaginal-free))
                  ((not (eq '= (caar action-spec)))
                   (print-warning "An imaginal-action simple-action request requires the = specifier for the action.")
                   (set-imaginal-free))
                  ((not (or (functionp (third (car action-spec))) (fboundp (third (car action-spec)))))
                   (print-warning "An imaginal-action simple-action request requires the action to name a valid function.")
                   (set-imaginal-free))
                  (t
                   (let ((c (funcall (third (car action-spec))))
                         (delay (if (imaginal-module-randomize instance)
                                    (randomize-time (imaginal-module-delay instance))
                                  (imaginal-module-delay instance))))
                     
                     (schedule-clear-buffer 'imaginal 0 :module 'imaginal)
                     (cond ((null c) ;; set module free and error t
                            (schedule-event-relative delay
                                                      'set-imaginal-free
                                                      :module 'imaginal
                                                      :output nil)
                            
                            (schedule-event-relative delay
                                                     'set-imaginal-error
                                                     :module 'imaginal
                                                     :output nil))
                           
                           ((chunk-p-fct c) ;; set module free and error t
                            (schedule-set-buffer-chunk 'imaginal c delay :module 'imaginal :priority -1000)
                            
                            (schedule-event-relative delay 'set-imaginal-free :module 'imaginal :output nil :priority -1001 :maintenance t))
                           (t
                            (model-warning "Invalid result from the action of an imaginal-action simple-action function.")
                            (set-imaginal-free))))))))
                            
                            
         (t (print-warning "Invalid request ~S to the imaginal-action buffer." (chunk-spec-chunk-type chunk-spec)))))))))

;;; Write a function that adds activation to an instruction to the extent that its conditions are satisfied
;;; :spreading-hook? or :activation-offsets --> allows adding additional components to activation.
;;; - Collect the conditions in the list of conditions
;;; - Check whether the condition is satisfied. If it is, add to the spreading.
;;; :spreading-hook it is. It takes 1 parameter, the chunk
;;; - If the chunk-type is not instr, then return 0.
;;; Otherwise, follow the condition chain to collect all conditions in a list.
;;; Once all conditions are collected, check whether they are true
;;; There are three types: nil tests, unequal tests, and equal test.
;;; In order to do this effectively, we should disect the test into its three components
;;; --> make it into a string and parse the string

;;; Spreading activation if conditions are satisfied addition
;;;

(defparameter *condition-spread* 3.0) ;;; extra activation when conditions are satisfied
(defparameter *condition-penalty* 0.0) ;;; activation punishment on a mismatch

(defun char-type (c)
  (cond
   ((digit-char-p c) 'digit)
   ((both-case-p c) 'letter)
   (t 'other)))

(defun parse-one-condition (chunk)
  (let ((s (string chunk))
        (result nil)
        (part ""))
    (loop while (not (equal s "")) do
          (when (and (not (equal part ""))
                     (not (eq (char-type (char part (1- (length part))))
                              (char-type (char s 0)))))
            (push part result)
            (setf part ""))
          (setf part (concatenate 'string part (subseq s 0 1)))
          (setf s (subseq s 1)))
    (push part result)
    (reverse result)))
    
(defun find-value-of (buffer-string slot-number instr-chunk)
   (if (or (equal buffer-string "CONTROL")(equal buffer-string "TASK")(equal buffer-string "PARENT")(equal buffer-string "GWM"))
      (let ((chunk (act-r-buffer-chunk (buffer-instance 'goal))))
        (when chunk
        	(cond  ((equal buffer-string "CONTROL") (chunk-slot-value-fct chunk 'control))
               ((equal buffer-string "PARENT") (chunk-slot-value-fct chunk 'parent))
               ((equal buffer-string "GWM")(chunk-slot-value-fct chunk 'gwm))
               (t (chunk-slot-value-fct chunk 'task)))))
    (let ((chunk 
           (cond
            ((equal buffer-string "CONST") instr-chunk)
            ((equal buffer-string "RT") (act-r-buffer-chunk (buffer-instance 'rharvest)))
            ((equal buffer-string "V") (act-r-buffer-chunk (buffer-instance 'perception)))
             ((equal buffer-string "PS") (act-r-buffer-chunk (buffer-instance 'imaginal)))
            (t (error "Illegal buffer-string ~S~%" buffer-string))))
          (slotname (second (assoc slot-number '(("0" id)("1" slot1)("2" slot2)("3" slot3)("4" slot4)("5" slot5)("6" slot6)) :test #'equal))))
      (when chunk (chunk-slot-value-fct chunk slotname)))
    ))

(defun test-condition (c chunk)

  (let* ((l (parse-one-condition c))
         (b1 (find-value-of (first l)(second l) chunk))
         (b2 (if (equal (fourth l) "NIL") nil 
               (find-value-of (fourth l)(fifth l) chunk)))
         (test (third l)))
    (cond ((and (equal (fourth l) "NIL") (equal test "=")) (equal b1 b2))
    	  ((equal test "=") (and b1 b2 (equal b1 b2)))
    	  ((and (equal test "<>") (equal (fourth l) "NIL")) b1)
          ((equal test "<>") (and b1 b2 (not (equal b1 b2)))))))


(defun spreading-condition (chunk)
  (if (eq (chunk-chunk-type-fct chunk) 'instr)
      (let ((condition-list nil)
            (next-condition (chunk-slot-value-fct chunk 'condition)))
        (loop while (not (null next-condition)) do
              (push (chunk-slot-value-fct next-condition 'c) condition-list)
              (setf next-condition (chunk-slot-value-fct next-condition 'cnext)))
        (let ((result t))
          (dolist (c condition-list)
            (when (and (not (eq c 'stop))(not (test-condition c chunk))) (setf result nil)))
        (when (and result (eq *verbose* 'full)) (format t "        ~8,2F: ~A receives spreading activation~%" (mp-time) chunk))

   
          (if result *condition-spread* *condition-penalty*)
        ))
    0.0))


;;; Extra utility for productions that complete a whole instruction
;;;
;;; This favors a productions that matches an instruction retrieval, and checks all conditions
;;; and carries out all actions. 
;;; Why did we put this in? Hack to make Stroop work?
;;;

(defun full-instruction-interpreter (production)
  (let* ((prod (produce-standard-representation production))
  	     (condition (first prod))
         (action (second prod))
         (goal-action (second (assoc '=goal> action)))
         (retrieval-condition (second (assoc '=retrieval> condition))))
    (and (member '(isa instr) retrieval-condition :test #'equal)
    (or (member '(action wait-all-done) goal-action :test #'equal)
        (member '(action nil) goal-action :test #'equal)))))
         


(defun extra-utility (production)
  (if (full-instruction-interpreter (get-production production)) 2.0 0.0))


;;; The model uses several extra buffers

;;; It uses a perception and action buffers to simplify communication with the outside world.
;;; The results of perception are places in the percetpion buffer
;;; Actions in the external world are +action> actions that are implemented
;;; by task-specific code.
;;; There may be a pending action if the action has a later effect. 
;;; Currently restricted to one for simplicity

(defstruct pmmodule stuffed busy pending-action change)

(defun create-pmmodule (model-name)
  (declare (ignore model-name))
  (make-pmmodule))

(defun reset-pmmodule (instance)
  (format t "Resetting module pmmodule~%")
  
  (sgp :do-not-harvest perception)
  (sgp :do-not-harvest action)
 (specify-compilation-buffer-type perception imaginal)  ;;; was goal instead of imaginal
 (specify-compilation-buffer-type action imaginal)  ;;; was goal instead of imaginal
 (setf (pmmodule-busy instance) nil)
 (setf (pmmodule-pending-action instance) nil)
 (setf (pmmodule-change instance) nil)
  )


(defun query-pmmodule (instance buffer-name slot value)
;  (format t "Query made of the pm module: ~S ~S ~S~%"  buffer-name slot value)
     (case slot
       (state
        (case value
          (busy
           (pmmodule-busy instance))
          (free
           (not (pmmodule-busy instance)))
          (error
              nil)
          (t
           (print-warning "Unknown state query ~S to pm module" value)
           nil))
        )
       (buffer
        (if (eql value 'stuffed)
            (pmmodule-stuffed instance)
            (print-warning "unknown buffer query ~S to pm module" value)
          ))
       (change
         (eq value (pmmodule-change instance)))  
       
       ))

(defun pending-action-p ()
  (if (pmmodule-pending-action (get-module pmmodule)) t nil))

(defun finalize-action (instance)
  (let* ((chunk-spec (eval `(define-chunk-spec isa fact slot1 ,(first *perception*)
                                               slot2 ,(second *perception*)
                                               slot3 ,(third *perception*)
                                               slot4 ,(fourth *perception*))))
         (chunk-action (eval `(define-chunk-spec isa fact))) )
    (setf (pmmodule-busy instance) nil)
  	(goal-style-request instance 'perception chunk-spec)
        (goal-style-request instance 'action chunk-action)
))


(defun finalize-perception (instance)
  (let* ((chunk-spec (eval `(define-chunk-spec isa fact slot1 ,(first *perception*)
                                               slot2 ,(second *perception*)
                                               slot3 ,(third *perception*)
                                               slot4 ,(fourth *perception*))))
          )
    (setf (pmmodule-busy instance) nil)
  	(goal-style-request instance 'perception chunk-spec)
  ))



(defun finalize-delayed-action (instance)
  (setf (pmmodule-change instance) t))

(defun schedule-delayed-action (percept latency)
  (setf *perception* '(pending))
  (setf (pmmodule-pending-action (get-module pmmodule)) percept)
  (schedule-event-relative latency #'finalize-delayed-action :params (list (get-module pmmodule)) :output 'medium))

(defun change-input (percept instance)
  (setf *perception* percept)
;  (finalize-action instance)
)

(defun schedule-delayed-percept (percept latency)
  "Also schedules a change in perception, but changes the buffer right away instead of just flagging the change"
  (schedule-event-relative latency #'change-input :params (list percept (get-module pmmodule))))


(defun request-pmmodule (instance buffer-name chunk-spec)
     (cond
     	((eq (chunk-spec-chunk-type chunk-spec) 'fact) (goal-style-request instance buffer-name chunk-spec))
        ((eq (chunk-spec-chunk-type chunk-spec) 'update-pmmodule)  ;;; respond to a delayed state change
  ;       (format t "~% *** Setting up pending action ~A~%" (pmmodule-pending-action instance))
         (setf (pmmodule-change instance) nil)
         (setf (pmmodule-busy instance) t)
         (setf *perception*  (pmmodule-pending-action instance))
         (setf (pmmodule-pending-action instance) nil)
         (schedule-event-relative 0 #'finalize-action :params (list instance) :output 'medium ))
        ((eq (chunk-spec-chunk-type chunk-spec) 'update-perception) 
  ;       (format t "~% *** Setting up pending action ~A~%" (pmmodule-pending-action instance))
         (when (not (pmmodule-pending-action instance))
           (setf (pmmodule-change instance) nil)
           (setf (pmmodule-busy instance) t)
           (setf (pmmodule-pending-action instance) nil))
         (schedule-event-relative 0 #'finalize-perception :params (list instance) :output 'medium ))

      ((eq buffer-name 'perception)(print-warning "Can only request perception with a fact"))
     	((eq (chunk-spec-chunk-type chunk-spec) 'external-action) 
         (let* ((action (if (slot-in-chunk-spec-p chunk-spec 'ac) 
                            (third (first 
                                    (chunk-spec-slot-spec chunk-spec 'ac))) 
                          nil))
                
                (arg1 (if (slot-in-chunk-spec-p chunk-spec 'slot1)
                          (third (first 
                                  (chunk-spec-slot-spec chunk-spec 'slot1)))
                        nil
                        ))
                
                (arg2 (if (slot-in-chunk-spec-p chunk-spec 'slot2)
                          (third (first 
                                  (chunk-spec-slot-spec chunk-spec 'slot2)))
                        nil
                        ))
                
                latency 
                )
           (cond ((eq action 'wait) ;;; Special case for wait
                  (when *verbose* (format t "*** ~8,2F: ACTION: Wait~%" (mp-time)))
                 (if (or (pmmodule-pending-action instance)(pmmodule-change instance)) (setf (pmmodule-busy instance) t)
                    (schedule-event-relative 0 #'finalize-action :params (list instance) :output 'medium ))
                  ;;; If there is a pending action, we set the state to busy so retrieve-instruction will wait for the pending action
                  ;;; If the state has already been changed, we switch state to busy as well, so that the appropriate production can pick up
                  ;;; Otherwise, wait will schedule the current perception into the state buffer
 ;                 (format t "*** ~%Doing a wait with action ~A and Statechange ~A~%" (pmmodule-pending-action instance)(pmmodule-change instance))
                  (setf latency 0))
                 (t
                  (when *verbose*
                  (cond ((and (null arg1)(null arg2))
                         (format t "*** ~8,2F: ACTION: ~A~%" (mp-time) action))
                        ((null arg2)
                         (format t "*** ~8,2F: ACTION: ~A ~A~%" (mp-time) action arg1))
                        (t
                         (format t "*** ~8,2F: ACTION: ~A ~A ~A~%" (mp-time) action arg1 arg2))))
                  (let ((tmp *perception*))
                    (setf latency (funcall (task-pm-function *task*) action arg1 arg2))
                    (when (and *verbose* (not (equal tmp *perception*))) (format t "*** ~8,2F: PERCEPTION: ~A~%" (mp-time) *perception*)))
                  (setf (pmmodule-busy instance) t)
                  (schedule-event-relative latency #'finalize-action :params (list instance) :output 'medium )))))

)


     )
     	   
;;; Because we build up retrievals one bit at a time, we need a buffer to build up the request,
;;; the rrequest buffer. The result of a retrieval is also used one bit at a time, and therefore
;;; we need a harvest buffer, rharvest

(defstruct rrequest stuffed)

(defun create-rrequest (model-name)
  (declare (ignore model-name))
  (make-rrequest))

(defun reset-rrequest (instance)
  (format t "Resetting module rrequest~%")
  
  (sgp :do-not-harvest rrequest)
 (specify-compilation-buffer-type rrequest goal)
  )


(defun query-rrequest (instance buffer-name slot value)
;  (format t "Query made of the rrequest module: ~S ~S ~S~%"  buffer-name slot value)
     (case slot
       (state
        (case value
          (busy
           nil)
          (free
           t)
          (error
              nil)
          (t
           (print-warning "Unknown state query ~S to rrequest module" value)
           nil))
        )
       (buffer
        (if (eql value 'stuffed)
            (rrequest-stuffed instance)
            (print-warning "unknown buffer query ~S to rrequest module" value)
          ))
       
       ))

(defstruct rharvest stuffed)

(defun create-rharvest (model-name)
  (declare (ignore model-name))
  (make-rharvest))

(defun reset-rharvest (instance)
  (format t "Resetting module rharvest~%")
  
  (sgp :do-not-harvest rharvest)
 (specify-compilation-buffer-type rharvest goal)
  )


(defun query-rharvest (instance buffer-name slot value)
;  (format t "Query made of the rharvest module: ~S ~S ~S~%"  buffer-name slot value)
     (case slot
       (state
        (case value
          (busy
           nil)
          (free
           t)
          (error
              nil)
          (t
           (print-warning "Unknown state query ~S to rharvest module" value)
           nil))
        )
       (buffer
        (if (eql value 'stuffed)
            (rharvest-stuffed instance)
            (print-warning "unknown buffer query ~S to rharvest module" value)
          ))
       
       ))

;;; Code for the imaginalreq buffer

(defstruct imaginalreq stuffed)

(defun create-imaginalreq (model-name)
  (declare (ignore model-name))
  (make-imaginalreq))

(defun reset-imaginalreq (instance)
  (format t "Resetting module imaginalreq~%")
  
  (sgp :do-not-harvest imaginalreq)
 (specify-compilation-buffer-type imaginalreq imaginal)
  )


(defun query-imaginalreq (instance buffer-name slot value)
;  (format t "Query made of the imaginalreq module: ~S ~S ~S~%"  buffer-name slot value)
     (case slot
       (state
        (case value
          (busy
           nil)
          (free
           t)
          (error
              nil)
          (t
           (print-warning "Unknown state query ~S to imaginalreq module" value)
           nil))
        )
       (buffer
        (if (eql value 'stuffed)
            (imaginalreq-stuffed instance)
            (print-warning "unknown buffer query ~S to imaginalreq module" value)
          ))
       
       ))




;;; The following code parses instructions 

(defvar *action-sequences*)
(setf *action-sequences* nil)
(defvar *condition-sequences*)
(setf *condition-sequences* nil)
(defvar *pm-function*)
(defvar *perception*)
(setf *perception* nil)
(defvar *all-tasks*)
(defstruct task name pm-function init-function reward parameters)
(setf *all-tasks* nil)
(defvar *descriptions*)
(defvar *vertices*)
(defvar *edges*)
(defvar *autoreorder*)
(setf *autoreorder* t)

(defmacro add-instr (&body l)
  `(instr-fct ',l))

(defmacro add-task (&body l)
  `(instr-fct ',l))

#|
(defun find-second (x l)
  (cond
   ((null l) nil)
   ((equal (second (first l)) x) (first (first l)))
   (t (find-second x (rest l)))))
|#

(defun permutep2 (l1 l2)
  (cond ((null l1) t)
        ((member (first l1) l2) (permutep2 (rest l1) l2))
        (t nil)))

(defun permutep (l1 l2)
  (if (= (length l1) (length l2)) (permutep2 l1 l2) nil))

(defun find-second (x l)
  (cond
   ((null l) nil)
   ((permutep (second (first l)) x) (first (first l)))
   (t (find-second x (rest l)))))



(defun parse-action-list (l)
 ; (push (list name l) *action-sequences*)
  (setf l (reverse l))
  (let ((l2 nil)
        (last-name 'astop)
        action)
    (loop while (not (null l)) do
        (push (first l) l2)
        (let ((new-action (find-second l2 *action-sequences*)))
          (if new-action (setf last-name new-action)
            (progn
              (setf action (new-name "AC"))
              (push (list action 'action (first l) 0) *vertices*)
              (when (not (eq last-name 'astop))(push (list action last-name 0) *edges*))
              (eval `(add-dm
                      (,action isa action name ,action a ,(first l) anext ,last-name)))
              (push (list action l2) *action-sequences*)
              (setf last-name action))))
        (setf l (rest l)))
    last-name))



(defun parse-condition-list (l)
 ; (push (list name l) *action-sequences*)
  (setf l (reverse l))
  (let ((l2 nil)
        (last-name 'cstop)
        condition)
    (loop while (not (null l)) do
        (push (first l) l2)
        (let ((new-condition (find-second l2 *condition-sequences*)))
          (if new-condition (setf last-name new-condition)
            (progn
              (setf condition (new-name "CD"))
              (push (list condition 'condition (first l) 0) *vertices*)
              (when (not (eq last-name 'cstop))(push (list condition last-name 0) *edges*))
              (eval `(add-dm
                      (,condition isa condition name ,condition c ,(first l) cnext ,last-name)))
              (push (list condition l2) *condition-sequences*)
              (setf last-name condition))))
        (setf l (rest l)))
    last-name))
    

(defun parse-instruction (task l)
  (let ((in-name (new-name "INSTR"))
        (condition nil) 
        (const1 'null)
        (const2 'null)
        (const3 'null)
        (const4 'null)
        (const5 'null)
        (const6 'null)
        (action nil))
   ; (push (list task task task) *vertices*)
    (loop while (not (null l)) do
          (cond
           ((eq (first l) ':condition)
            (pop l)
            (if (not (setf condition (find-second (first l) *condition-sequences*)))
                (setf condition (parse-condition-list (pop l)))
              (pop l)))
          ((eq (first l) ':action)
            (pop l)
            (if (not (setf action (find-second (first l) *action-sequences*)))
              (setf action (parse-action-list (pop l)))
              (pop l)))
          ((eq (first l) ':const)
           (pop l)
           (when (first (first l)) (setf const1 (first (first l))))
           (when (second (first l)) (setf const2 (second (first l))))
           (when (third (first l)) (setf const3 (third (first l))))
           (when (fourth (first l)) (setf const4 (fourth (first l))))
           (when (fifth (first l)) (setf const5 (fifth (first l))))
           (when (sixth (first l)) (setf const6 (sixth (first l))))
           (pop l))
          ((eq (first l) ':description)
           (pop l)
           (push (list in-name (pop l)) *descriptions*))

          (t (format t "Cannot parse ~A~%" (pop l)))))
    (push (list in-name task in-name 0) *vertices*)
    (push (list in-name condition 0) *edges*)
    (when (not (eq action 'astop))  (push (list in-name action 0) *edges*))
;;; Extra
    (push (list task in-name 1) *edges*)

    (eval `(add-dm (,in-name isa instr task ,task condition ,condition action ,action slot1 ,const1 slot2 ,const2 slot3 ,const3 slot4 ,const4 slot5 ,const5 slot6 ,const6)))))


(defun find-constants (l transtable)
	(let ((result nil))
		(dolist (x l result)
                  (when (eq x 'x) (push 'x result))
			(when (listp x)
			(when 
				(and (first x)  ;;; not nil
				     (not (member (first x) '(RT1 RT2 RT3 RT4 newWM newWM1 newWM2 newWM3 newWM4 RT PS1 PS2 PS3 PS4 AC1 AC2 AC3 control0 task0 done ))) ;; not one of the reserved words
                                     (not (member (first x) result)) ;;; not already on the list
                                     (not (assoc (first x) transtable))) ;;; not already used as a variable
                          (push (first x) result))
                          (when 
                              (and (third x)  ;;; not nil
                                   (not (member (third x) '(RT1 RT2 RT3 RT4 newWM newWM1 newWM2 newWM3 newWM4 RT PS1 PS2 PS3 PS4 AC1 AC2 AC3 control0 task0 done ))) ;; not one of the reserved words
                                   (not (member (third x) result)) ;;; not already on the list
                                   (not (assoc (third x) transtable))) ;;; not already used as a variable
			    (push (third x) result))))
		result))
			    
(defconstant *condition-order* '(V1 V2 V3 V4 PS0 PS1 PS2 PS3 PS4 control0 task0 gwm0 parent0 RT0 RT1 RT2 RT3 RT4 CONST1 CONST2 CONST3 CONST4 CONST5 CONST6))

(defun find-position (x &optional (l *condition-order*)(pos 1))
  (cond ((null l) 100)
        ((eq x (first l)) pos)
        (t (find-position x (cdr l) (1+ pos)))))

(defun translate-ca (h consts transtable)
   (if (atom h) h
  (let* ((tt (append (merge-lists consts '(CONST1 CONST2 CONST3 CONST4 CONST5 CONST6)) transtable))
  	     (buffer1 (second (assoc (first h) tt)))
  	     (buffer2 (when (third h) (second (assoc (third h) tt)))))
  	     (when (null buffer1) (setf buffer1 (first h)))
  	     (when (null buffer2) (setf buffer2 (third h)))
    (if (and (member (second h) '(= <>)) buffer2 (> (find-position buffer1) (find-position buffer2)))
      (intern (format nil "~A~A~A" buffer2 (second h) buffer1))
    (intern (format nil "~A~A~A" buffer1 (second h) buffer2))))))

(defun find-operator (x &optional (pos 0)) 
  (cond 
   ((< (length x) 1) nil)
   ((equal (subseq x 0 1) "=") (list pos '=))
   ((< (length x) 2) nil)
   ((equal (subseq x 0 2) "<>") (list pos '<>))
   ((equal (subseq x 0 2) "->") (list pos '->))
   (t (find-operator (subseq x 1) (1+ pos)))))

(defun parse-single (x)
  "Break up symbol in subparts, if necessary"
  (if (or (listp x) (member x '(= <> ->))) (list x)
    (let* ((stringx (format nil "~A" x))
           (op (find-operator stringx))
           (endpos (when op (+ (first op) (if (eq (second op) '=) 1 2)))))
      (if op
          (let* ((item1 (when (> (first op) 0) (intern (subseq stringx 0 (first op)))))
                 (res (when (not (= (length stringx) endpos))(list (intern (subseq stringx endpos))))))
            (push (second op) res)
            (when item1 (push item1 res))
            res)
        (list x)))))

(defun parse-out-singles (l)
  (cond ((null l) nil)
        (t (append (parse-single (first l))(parse-out-singles (rest l))))))

(defun parse-three (l)
	"Group elements in groups of three, and parse out any retrieval, action and episodic lists"
          (cond 
           ((null l) nil)
            ((eq (first l) 'done) '(done))
            ((and (listp (first l)) (eq (third l) 'AC))
             (append (reverse (merge-lists (first l) '(AC1 AC2 AC3) (second l))) (parse-three (cdddr l))))
            ((and (listp (first l)) (eq (third l) 'newWM))
             (append (reverse (merge-lists (first l) '(newWM1 newWM2 newWM3 newWM4) (second l))) (parse-three (cdddr l))))
            ((and (listp (first l)) (eq (third l) 'RT))
             (append (reverse (merge-lists (first l) '(RT1 RT2 RT3 RT4) (second l))) (parse-three (cdddr l))))		 
            (t (cons (list (first l)(second l)(third l)) (parse-three (cdddr l))))))

(defun add-x-to-list (l)
  (cond ((null l) nil)
        ((eq (first l) 'x) (cons 'x l))
        (t (cons (first l)(add-x-to-list (rest l))))))

(defun parse-instruction-v (task l transtable declarative)
   (let (
        (condition nil)
        (consts nil)
        (action nil)
        (description ""))
    (loop while (not (null l)) do
          (cond
           ((eq (first l) ':condition)
            (pop l)
            (setf condition (parse-three (parse-out-singles (pop l) ))))
          ((eq (first l) ':action)
            (pop l) 
            (setf action (parse-three (parse-out-singles (pop l) ))))
          ((eq (first l) ':description)
           (pop l)
           (setf description (pop l)))
          (t (format t "~%Unknown keyword ~A~%" (first l))
             (pop l))
           ))
	(setf consts (find-constants (append (reverse action) '(x) condition) transtable)) ;;; Changed this! Potentially radical reversal of action & condition
        ;;; Now blow up the x so that the length of the list is exactly 6.
        (when (> (length consts) 7) (format t "~%Too many constants in Instruction~%"))
    (if (> (length consts) 6) (setf consts (remove 'x consts))
      (loop while (< (length consts) 6) do
           (setf consts (add-x-to-list consts))))
    (setf consts (substitute 'null 'x consts))
    (setf condition (mapcar #'(lambda (x) (translate-ca x consts transtable)) condition))

    (setf action  (mapcar #'(lambda (x) (translate-ca x consts transtable)) action))
    (parse-instruction task (list ':condition condition ':action action ':const consts ':description description))))

(defun parse-prod (l transtable)
  (let ((result nil)
        (current-buffer nil)
        (temp nil)
        (neg nil))
    (loop while (not (null l)) do
      (cond ((member (first l) '(=input> =imaginal> =goal> =action> =retrieval> +action> +retrieval>)) (setf current-buffer (pop l)))
            ((eq (first l) '-) (setf neg t)(pop l))
            ((eq (first l) '==>) (setf result temp temp nil) (pop l))
            ((assoc (first l) transtable) ;; no checking yet whether buffer is correct. Also check for the literal version, e.g. RT1
             (push (list neg (pop l) (pop l)) temp)
             (setf neg nil))
            (t (format t "Illegial expression ~A~%" (pop l)))
      
      ))
    (list result temp)))


(defun parse-classic-instruction (task l transtable declarative)
  (let* ((condition nil)
         (action nil)
         (description (pop l))
         (prod (parse-prod l transtable))
         (lhs (reverse (first prod)))
         (rhs (reverse (second prod)))
         
         )
    (format t "Parsing rule ~A~%" description)
    (loop while (not (null lhs)) do
      (let ((x (first lhs)))
        (cond ((null (third x)) ; A nil test of a slot
               (setf condition (append condition (list (second x) (if (null (first x)) '= '<>) nil)))
               (pop lhs))
              ((not (chunk-spec-variable-p (third x)))  ; Matching a constant in a slot
               (setf condition (append condition (list (second x) (if (null (first x)) '= '<>) (third x))))
               (pop lhs))
              ((null (first x)) ;; There is non-negated variable in the slot
               (let ((temp nil))  ;; There is a variable in the slot, so find all matching lhs variables
                 (dolist (y (rest lhs)) (when (eq (third x)(third y)) (push y temp)))
                 (dolist (y (reverse temp)) ;; now add all pairs to the condition. If the list is empty, no conditions are added
                   (setf condition (append condition (list (second x) (if (null (first y)) '= '<>) (second y))))
                   (setf lhs (remove y lhs :test 'equal)))
                 (pop lhs)))
              (t ;;; The only possibility left is a negated variable. We'll whine about this one
               (format t "Illegal test ~A~%" x)
               (pop lhs)))))
    (format t "  Condition: ~A~%" condition)
    (setf lhs (reverse (first prod)))
    (dolist (x rhs)  ;; action are more straightforward
      (cond ((chunk-spec-variable-p (third x))  ;; Variable
             (let ((temp nil)
                   (lhs (reverse (first prod))))
               (loop while (and lhs (null temp)) do
                 (when (and (null (first (first lhs))) (eq (third (first lhs)) (third x)))
                   (setf temp (second (first lhs))))
                 (pop lhs))
               (if temp (setf action (append action (list temp '-> (second x))))
                 (format t "Non-matching rhs ~A~%" x))))
            ((not (null (third x))) (setf action (Append action (list (third x) '-> (second x)))))
            (t (format t "Can't set a slot to nil in ~A~%" x))))
     (format t "  Action   : ~A~%" action)
   (parse-instruction-v task (list ':condition condition ':action action ':description description) transtable declarative)
  ))
    

(defun merge-lists (l1 l2 &optional cnst)
  "Multi-purpose list merger: can insert an constant in each list, and removes ?"
	(cond
		((or (null l1) (null l2)) nil)
		((eq (first l1) '?) (merge-lists (rest l1)(rest l2) cnst))
		(cnst (cons (list (first l1) cnst (first l2)) (merge-lists (rest l1)(rest l2) cnst)))
		(t (cons (list (first l1)(first l2)) (merge-lists (rest l1)(rest l2))))))

(defun flatten (l)
  "flatten list: merge sublists into main and remove nils"
  (cond ((null l) nil)
        ((null (first l)) (flatten (rest l)))
        ((listp (first l)) (flatten (append (first l) (rest l))))
        (t (cons (first l)(flatten (rest l))))))
		

(defun instr-fct (l)
  (let ((task-name (pop l)) 
         (transtable '((Gtask task0)(Gcontrol control0)(Gtop gwm0)(Gparent parent0)(ACaction AC1)(ACarg1 AC2)(ACarg2 AC3)
                       (RTid RT0)(WMid PS0)(PSid PS0)
                       ))
  declarative
     (task (make-task))
  	)
  	(setf (task-name task) task-name)
        (push (list task-name task-name task-name 1) *vertices*)
  	(loop while (atom (first l)) do
              (cond
               ((eq (first l) ':input)
                (setf transtable (append (merge-lists (second l) '(V1 V2 V3 V4)) transtable))
                (pop l)(pop l))
               ((member (first l) '(:variables :working-memory))
                (setf transtable (append (merge-lists (second l) '(PS1 PS2 PS3 PS4)) transtable))
                (pop l)(pop l))

               ((eq (first l) ':declarative)
                (setf declarative (second l))
                (dolist (x declarative)
                  (setf transtable (append (merge-lists x '(RT1 RT2 RT3 RT4)) transtable)))
                (pop l)(pop l))
               ((eq (first l) ':reward)
                (pop l)
                (setf (task-reward task) (pop l)))
               ((eq (first l) ':pm-function)
                (pop l)
                (setf (task-pm-function task) (pop l)))
               ((eq (first l) ':init)
                (pop l)
                (setf (task-init-function task) (pop l)))
               ((eq (first l) ':facts) (pop l)
                (eval `(add-dm ,@(pop l))))
               ((eq (first l) ':parameters) (pop l)
                (setf (task-parameters task)(pop l)))
               (t (pop l)(pop l)))) 
  	(push task *all-tasks*)
    (dolist (x l)
      (cond
       ((eq (first x) 'in)
        (parse-instruction task-name (rest x))
       )
       ((member (first x) '(ins operator op))
        (parse-instruction-v task-name (rest x) transtable declarative)
        )
       ((eq (first x) 'p) 
        (parse-classic-instruction task-name (rest x) transtable declarative))
))))
        
(defun set-task (task-name)
  (let ((task (find task-name *all-tasks* :test #'(lambda (x y)(eq x (task-name y))))))
  	(setf *task* task)
  ;	(eval `(spp action-done :reward ,(task-reward task)))
  	(when (task-parameters task) (dolist (x (task-parameters task)) (eval x)))
  	))
  	
(defun init-task ()
  (clear-buffer 'goal)
  (funcall (task-init-function *task*))

  (when *perception* 
    (schedule-event-relative 0 #'finalize-action :params (list (get-module pmmodule)) :output 'medium ))


;    	(eval `(spp action-done :reward ,(task-reward *task*)))

 ; 	(when (task-parameters *task*) (dolist (x (task-parameters *task*)) (eval x)))
)

(defun issue-reward ()
  (trigger-reward  (task-reward *task*))
)

(defvar *verbose* nil)

(defun trace-output (s)
  (when *verbose*
  (format t "~A~%" s)))


(defun trace-retrieved-instructions (chunk)

(when (and *verbose* chunk (eq (chunk-chunk-type-fct chunk) 'instr))
  
  (format t "*** ~8,2F: ~A: ~A~%" (mp-time) chunk (second (assoc chunk *descriptions*))))
(when (and (eq *verbose* 'full) chunk (eq (chunk-chunk-type-fct chunk) 'condition))
  (let ((value (chunk-slot-value-fct chunk 'c)))
    (if (eq value 'stop)
        (format t "        ~8,2F: All conditions matched~%" (mp-time))
      (format t "        ~8,2F: Testing condition ~A: ~A~%" (mp-time) chunk value))))
(when (and (eq *verbose* 'full) chunk (eq (chunk-chunk-type-fct chunk) 'action))
  (let ((value (chunk-slot-value-fct chunk 'a)))
    (if (eq value 'stop)
        (format t "        ~8,2F: All actions done~%" (mp-time))
      (format t "        ~8,2F: Carrying out action ~A: ~A~%" (mp-time) chunk value))))
(when (and *verbose* chunk (eq (chunk-chunk-type-fct chunk) 'fact))
      (format t "*** ~8,2F: Retrieving fact ~A: ~A ~A ~A ~A~%" (mp-time) chunk (chunk-slot-value-fct chunk 'slot1)
              (chunk-slot-value-fct chunk 'slot2)(chunk-slot-value-fct chunk 'slot3)(chunk-slot-value-fct chunk 'slot4)))
  
)



;;; Print out nodes and vertices in order to plot a network of instruction DM
;;; Used in combination with an R script
;;;

(defun print-net ()
  ; Vertices first
  (with-open-file (f "~/vertices.txt" :direction :output :if-exists :overwrite :if-does-not-exist :create)
  (format f "name task label task-node~%")
  
  (dolist (x *vertices*) (format f "~10S ~10S ~10S ~10S~%" (first x)(second x)(third x)(fourth x))))
  (with-open-file (f "~/edges.txt" :direction :output :if-exists :overwrite :if-does-not-exist :create)

  (format f "from to task-node~%")
  (dolist (x *edges*)(format f "~10S ~10S ~10S~%" (first x)(second x)(third x)))))

;;; Now define a wrapper macro to define a new model

(defmacro define-model-transfer (&body model-code)
`(progn
(clear-all)

;(defmethod device-speak-string ((win rpm-window) text)

;(format t "~S~%" text)

;)
(setf *vertices* nil *edges* nil)
(setf *descriptions* nil)


(undefine-module imaginal)

(define-module-fct 'imaginal
    '(imaginal imaginal-action)
  (list
   (define-parameter :imaginal-delay :valid-test #'nonneg 
     :default-value .2 :warning "non-negative number" 
     :documentation "Time in seconds to respond to an imaginal request")
   (define-parameter :vidt :valid-test #'tornil :default-value nil
          :warning "T or nil" 
          :documentation "Variable Imaginal Delay Time"))
   
  :version "1.2"
  :documentation "The imaginal module provides a goal style buffer with a delay and an action buffer for manipulating the imaginal chunk"
  :creation #'create-imaginal
  :query #'imaginal-query
  :request #'imaginal-request-2
  :buffer-mod #'imaginal-mod-request
  :params #'imaginal-params
  :reset #'imaginal-reset)



;;; Add the new modules to ACT-R
(undefine-module pmmodule)          
(define-module-fct 'pmmodule '((perception nil nil (change)) action) nil
  :version "1.0a1"
  :documentation "pm module"
  :creation #'create-pmmodule
  :reset (list nil #'reset-pmmodule)
  :query #'query-pmmodule
  :request #'request-pmmodule)

(undefine-module rrequest)          
(define-module-fct 'rrequest '(rrequest) nil
  :version "1.0a1"
  :documentation "rrequest module"
  :creation #'create-rrequest
  :reset (list nil #'reset-rrequest)
  :query #'query-rrequest
  :request #'goal-style-request)

(undefine-module rharvest)
(define-module-fct 'rharvest '(rharvest) nil
  :version "1.0a1"
  :documentation "rharvest module"
  :creation #'create-rharvest
  :reset (list nil #'reset-rharvest)
  :query #'query-rharvest
  :request #'goal-style-request)

(undefine-module imaginalreq)
(define-module-fct 'imaginalreq '(imaginalreq) nil
  :version "1.0a1"
  :documentation "imaginalreq module"
  :creation #'create-imaginalreq
  :reset (list nil #'reset-imaginalreq)
  :query #'query-imaginalreq
  :request #'goal-style-request)



(define-model actransfer
(sgp :do-not-harvest imaginal :save-buffer-trace t :esc t :rt -0.5 :lf 0.5 :egs .03 :ul t :epl t :DECLARATIVE-NUM-FINSTS 300
     :DECLARATIVE-FINST-SPAN 3000.0 :alpha 0.1 :ut -10000 :pct t :er t :iu 10.0 :ans 0.1 :bll 0.5 :v nil
;     :utility-offsets extra-utility
) 
(setf *standard-motor-time* 0.5)
(setf *standard-visual-time* 0.5)
(setf *condition-spread* 0.3)  
(setf *condition-penalty* -0.3)
(setf *all-tasks* nil)

(sgp :activation-offsets spreading-condition)
(sgp :retrieved-chunk-hook trace-retrieved-instructions)

(chunk-type goal task condition action slot1 slot2 slot3 slot4 slot5 slot6 retrieval-request action-request episodial-request harvest-request control parent gwm)
(chunk-type instr task condition action slot1 slot2 slot3 slot4 slot5 slot6)
(chunk-type condition name c cnext slot1)
(chunk-type action name a anext slot1)
(chunk-type fact id slot1 slot2 slot3 slot4)
(chunk-type external-action ac slot1 slot2)
(chunk-type update-pmmodule)
(chunk-type update-perception)
(setf *action-sequences* nil)
(setf *condition-sequences* nil)
(setf *verbose* t)
(setf *descriptions* nil)
(setf *vertices* nil)
(setf *edges* nil)

(add-dm 

 (cstop isa condition name cstop c stop)
 (astop isa action name astop a stop)
)

,@model-code

(when (and (no-output (sgp :bll)) (< (caar (no-output (sdp cstop :reference-count))) 2))
(sdp :reference-count 1000 :creation-time -1000000)) ;; 11.5 days, only set if bll is on and it has not yet been set by the model
	  ;;; This will set the activation to approximately 0.7


(p start-a-goal
   ?goal>
     buffer empty
   ?imaginal>
     state free
   ?perception>
     buffer empty
==>
  !bind! =task (task-name *task*)
  +goal>
     isa goal
     task =task
;     condition pending
  +perception>
     isa fact
  +imaginal>
     isa fact
;     id pending
   +rharvest>
     isa fact
   +rrequest>
     isa fact
   +action>
     isa fact
   +imaginalreq>
     isa fact
   -retrieval>
  !safe-eval! (setf (dm-finsts (get-module declarative)) nil) ;;; Set it up for the next cycle   
  !eval! (setf (dm-failed (get-module declarative)) nil)
)

(p start-a-goal-with-initial-perception
   ?goal>
     buffer empty
   ?imaginal>
     state free
   ?perception>
     - buffer empty
==>
  !bind! =task (task-name *task*)
  +goal>
     isa goal
     task =task
;     condition pending
  +imaginal>
     isa fact
;     id pending
   +rharvest>
     isa fact
   +rrequest>
     isa fact
   +action>
     isa fact
   +imaginalreq>
     isa fact
  -retrieval>
  !safe-eval! (setf (dm-finsts (get-module declarative)) nil) ;;; Set it up for the next cycle   
  !eval! (setf (dm-failed (get-module declarative)) nil)
)

;;; All the buffer checks should drop out on compilation (assuming they do not issue new +buffers>, but that should be check by rule 2).

(p retrieve-instruction
   =goal>
     isa goal
     task =task
     condition nil
     action nil
   ?retrieval>
     state free
     buffer empty
   ?imaginal>
     state free
   ?action>
     state free
;   ?perception>
;     state free
  !safe-eval! (and (not (pmmodule-change (get-module pmmodule)))(not (pmmodule-busy (get-module pmmodule)))) 
;; same as ?perception> state free, but that was dropped during compilation
   !safe-eval! (not (dm-failed (get-module declarative)))
==>
  =goal>
    action pending
   +retrieval>
     isa instr
     task =task
     :recently-retrieved nil
   )

(spp retrieve-instruction :u 0.5)

(p retry-on-pmmodule-change
   =goal>
     isa goal
     condition nil
     action nil
   ?retrieval>
     state free
     buffer empty
   ?imaginal>
     state free
   ?perception>
     change t
==>
   +perception>
     isa update-pmmodule
  !safe-eval! (setf (dm-finsts (get-module declarative)) nil) ;;; Set it up for the next cycle   
)


(p retry-on-failure
   =goal>
     isa goal
     action pending
     condition nil
   ?retrieval>
      state error
==>
   !eval! (setf (dm-failed (get-module declarative)) nil)
  !safe-eval! (setf (dm-finsts (get-module declarative)) nil) ;;; Set it up for the next cycle   
  !safe-eval!  (trace-output (format nil "*** ~6,2F: Failed to retrieve an instruction, retrying." (mp-time))))



(p store-retrieved-instruction
   =goal>
     isa goal
     action pending
   =retrieval>
     isa instr
     condition =cond
     action =act
     slot1 =const1
     slot2 =const2
     slot3 =const3
     slot4 =const4
     slot5 =const5
     slot6 =const6
==>
  =goal>
    condition =cond
    action =act
     slot1 =const1
     slot2 =const2
     slot3 =const3
     slot4 =const4
     slot5 =const5
     slot6 =const6
  +retrieval>
    isa condition
    name =cond
;  !safe-eval!  (trace-output (format nil "*** ~6,2F: ~A: ~S" (mp-time) (chunk-copied-from-fct =retrieval) (second (assoc  (chunk-copied-from-fct =retrieval) *descriptions*))))
)

(spp store-retrieved-instruction :u 0.5)

(p update-perception-on-wait
   =goal>
     isa goal
     condition nil
     action nil
   ?retrieval>
     state free
     buffer empty
   ?imaginal>
     state busy
   ?action>
     state free
   !eval! t
==>
  +perception>
    isa update-perception)

(spp update-perception-on-wait :u 0.5)


;;; Loop through the nil checks

(defvar *problem-states*)
(defvar *retrieval-request-slots*)
(defvar *retrieval-harvest-slots*)
(defvar *const-slots*)
(defvar *visual-slots*)
(defvar *motor-slots*)
(defvar *imaginalreq-slots*)
(defvar *goal-slots*)

(setf *problem-states* '((PS0 id imaginal)(PS1 slot1 imaginal) (PS2 slot2 imaginal) (PS3 slot3 imaginal) (PS4 slot4 imaginal)))
(setf *retrieval-request-slots* '((RT0 id rrequest)(RT1 slot1 rrequest) (RT2 slot2 rrequest) (RT3 slot3 rrequest) (RT4 slot4 rrequest)))
(setf *retrieval-harvest-slots* '((RT0 id rharvest)(RT1 slot1 rharvest) (RT2 slot2 rharvest) (RT3 slot3 rharvest) (RT4 slot4 rharvest)))
(setf *const-slots* '((CONST1 slot1 goal)(CONST2 slot2 goal)(CONST3 slot3 goal)(CONST4 slot4 goal)(CONST5 slot5 goal)(CONST6 slot6 goal)))
(setf *visual-slots* '((V1 slot1 perception)(V2 slot2 perception)(V3 slot3 perception)(V4 slot4 perception)))
(setf *motor-slots* '((AC1 slot1 action)(AC2 slot2 action)(AC3 slot3 action)))  ;; removed (VOC string vocal)
(setf *imaginalreq-slots* '((newWM1 slot1 imaginalreq)(newWM2 slot2 imaginalreq)(newWM3 slot3 imaginalreq)(newWM4 slot4 imaginalreq)))
(setf *goal-slots* '((task0 task goal2)(control0 control goal2)(parent0 parent goal2)(gwm0 gwm goal2)))

(defvar *all-lhs-slots*)
(defvar *all-rhs-slots*)
(setf *all-lhs-slots* (append *problem-states* *const-slots* *visual-slots* *retrieval-harvest-slots* *goal-slots*))
(setf *all-rhs-slots* (append *problem-states* *retrieval-request-slots* *motor-slots* *imaginalreq-slots* *goal-slots*))

  (dolist (x *all-lhs-slots*)
    (let ((test (intern (format nil "~A=NIL" (first x))))
          (goal-test (if (eq (third x) 'goal) (list (second x) 'null)
          	(when (eq (third x) 'goal2) (list (second x) nil))))
          (imaginal-test (when (eq (third x) 'imaginal) (list '=imaginal> 'isa 'fact (second x) nil)))
          (retrieval-test (when (eq (third x) 'rharvest) (list '=rharvest> 'isa 'fact (second x) nil)))
          (pmmodule-test (when (eq (third x) 'perception)(list '=perception> 'isa 'fact (second x) nil))))
      (eval `(add-dm (,test isa chunk)))
    (eval
`(p ,(new-name-fct (format nil "check-~A=nil-" (first x)))
    =goal>
      isa goal
      ,@goal-test
      ,@retrieval-test
   =retrieval>
     isa condition
     c ,test
     cnext =c2
   ,@imaginal-test
   ,@pmmodule-test
==> 
   +retrieval>
    isa condition
    name      =c2))))


  (dolist (x *all-lhs-slots*)
    (let ((test (intern (format nil "~A<>NIL" (first x))))
          (goal-test (if (eq (third x) 'goal) (list (second x) '=value '!safe-eval! '(not (equal =value 'null)))
          	(when (eq (third x) 'goal2) (list (second x) '=value))))
          (imaginal-test (when (eq (third x) 'imaginal) (list '=imaginal> 'isa 'fact (second x) '=value)))
          (retrieval-test (when (eq (third x) 'rharvest) (list '=rharvest> 'isa 'fact (second x) '=value)))
          (pmmodule-test (when (eq (third x) 'perception)(list '=perception> 'isa 'fact (second x) '=value))))
      (eval `(add-dm (,test isa chunk)))
    (eval
`(p ,(new-name-fct (format nil "check-~A<>nil-" (first x)))
    =goal>
      isa goal
      ,@goal-test
      ,@retrieval-test
   =retrieval>
     isa condition
     c ,test
     cnext =c2
   ,@imaginal-test
   ,@pmmodule-test
==> 
   +retrieval>
    isa condition
    name      =c2))))

;;; Check-not-equal and check-equal
;;; Needs to loop through the following slots:
;;; PS1-4, RT1-4, CONST1-2, and later possibly others (visual, etc.)
;;;

(dolist (x *all-lhs-slots*)
      (let (
          (goal-test-1 (when (member (third x) '(goal goal2)) (list (second x) '=value1)))
          (imaginal-test-1 (when (eq (third x) 'imaginal) (list '=imaginal> 'isa 'fact (second x) '=value1)))
          (retrieval-test-1 (when (eq (third x) 'rharvest) (list '=rharvest> 'isa 'fact (second x) '=value1)))
          (pmmodule-test-1 (when (eq (third x) 'perception)(list '=perception> 'isa 'fact (second x) '=value1))))
        (dolist (y *all-lhs-slots*)
          (when (not (equal x y))
      (let ((test (intern (format nil "~A=~A" (first x)(first y))))
            (test-n (intern (format nil "~A<>~A" (first x)(first y))))
          (goal-test-2 (when (member (third y) '(goal goal2)) (list (second y) '=value2)))
          (imaginal-test-2 (when (eq (third y) 'imaginal) 
                             (if (null imaginal-test-1) (list '=imaginal> 'isa 'fact (second y) '=value2)
                                       (list (second y) '=value2))))
          (pmmodule-test-2 (when (eq (third y) 'perception) 
                             (if (null pmmodule-test-1) (list '=perception> 'isa 'fact (second y) '=value2)
                                       (list (second y) '=value2))))
          (retrieval-test-2 (when (eq (third y) 'rharvest) 
                             (if (null retrieval-test-1) (list '=rharvest> 'isa 'fact (second y) '=value2)
                                       (list (second y) '=value2)))))      
        
      (eval `(add-dm (,test isa chunk)(,test-n isa chunk)))
    (eval
`(p ,(new-name-fct (format nil "check-~A=~A-" (first x)(first y)))
    =goal>
      isa goal
      ,@goal-test-1
      ,@goal-test-2
      ,@retrieval-test-1
      ,@retrieval-test-2
      ,@pmmodule-test-1
      ,@pmmodule-test-2
   =retrieval>
     isa condition
     c ,test
     cnext =c2
   ,@imaginal-test-1
   ,@imaginal-test-2
   !safe-eval! (equal =value1 =value2)
==>
   ,@(when (or retrieval-test-1 retrieval-test-2) '(=goal> harvest-request t))
   +retrieval>
    isa condition
    name      =c2))
   (eval
`(p ,(new-name-fct (format nil "check-~A<>~A-" (first x)(first y)))
    =goal>
      isa goal
      ,@goal-test-1
      ,@goal-test-2
      ,@retrieval-test-1
      ,@retrieval-test-2
      ,@pmmodule-test-1
      ,@pmmodule-test-2
   =retrieval>
     isa condition
     c ,test-n
     cnext =c2
   ,@imaginal-test-1
   ,@imaginal-test-2
   !safe-eval! (not (equal =value1 =value2))
==>
   ,@(when (or retrieval-test-1 retrieval-test-2) '(=goal> harvest-request t))
   +retrieval>
    isa condition
    name      =c2))

)))))

(p general-condition-does-not-match
   =goal>
     isa goal
   =retrieval>
     isa condition
   !eval! t ;; compilation blocker
==> 
  !safe-eval! (trace-output (format nil "*** ~8,2F: Mismatch on condition ~A" (mp-time) =retrieval))

  =goal>
     condition nil
     action nil
     slot1 nil)

(p general-action-does-not-apply
   =goal>
     isa goal
   =retrieval>
     isa action
     a V1->RT2
   !eval! t ;; compilation blocker
==> 
  !safe-eval! (trace-output (format nil "*** ~8,2F: Action cannot be carried out ~A" (mp-time) =retrieval))
  !safe-eval! (setf (dm-finsts (get-module declarative)) nil) ;;; Set it up for the next cycle   
  =goal>
     condition nil
     action nil
     slot1 nil)


;(spp general-condition-does-not-match :reward 0)

;;;

(p check-condition-done
   =goal>
     isa goal
   =retrieval>
     isa condition
     c stop
==>
   =goal>
     condition act
)  

(p start-action-sequence
   =goal>
     isa goal
     condition act
     action =a1
   ?retrieval>
     state free
     buffer empty
   !safe-eval! (not (dm-failed (get-module declarative)))
==>
  =goal>
    condition nil
   +retrieval>
    isa action
    name      =a1

)  

(spp start-action-sequence :u 2.0)

;(spp check-done :reward 2.0)

;;; Action productions all move something from one slot to another
;;; Some slots are destination only, mainly motor actions

(dolist (x *all-lhs-slots*)
      (let (
          (goal-test (when (member (third x) '(goal goal2)) (list (second x) '=value)))
          (imaginal-test-1 (when (eq (third x) 'imaginal) (list '=imaginal> 'isa 'fact (second x) '=value)))
          (retrieval-test (when (eq (third x) 'rharvest) (list '=rharvest> 'isa 'fact (second x) '=value)))
          (retrieval-harvest (when (eq (third x) 'rharvest) '(=goal> harvest-request t)))
          (visual-test (when (eq (third x) 'perception)(list '=perception> 'isa 'fact (second x) '=value))))
        (dolist (y *all-rhs-slots*)
          (when (not (equal x y))
      (let ((test (intern (format nil "~A->~A" (first x)(first y))))
          (goal-action (let ((temp (cond ((member (third y) '(goal goal2)) (list (second y) '=value))
          					 ((eq (third y) 'rrequest) '(retrieval-request t))
                                                 ((eq (third y) 'imaginalreq) '(episodial-request t))
          					 ((eq (third y) 'action) '(action-request t))
                                                 )))
                         (cond ((and temp retrieval-harvest) (append retrieval-harvest temp))
                               (retrieval-harvest retrieval-harvest)
                               (temp (cons '=goal> temp)))))
          (imaginal-test-2 (when (and (null imaginal-test-1) (eq (third y) 'imaginal)) '(=imaginal> isa fact)))
          (imaginal-action (when (eq (third y) 'imaginal) 
                              (list '=imaginal> (second y) '=value)))
          (imaginalreq-test (when (eq (third y) 'imaginalreq) '(=imaginalreq> isa fact)))
          (imaginalreq-action (when (eq (third y) 'imaginalreq) (list '=imaginalreq> (second y) '=value)))
          (external-action-test (when  (eq (third y) 'action) '(=action> isa fact)))
          (external-action (when (eq (third y) 'action) (list '=action> (second y) '=value)))
          (retrieval-match (when (eq (third y) 'rrequest) '(=rrequest> isa fact)))                
          (retrieval-action (when (eq (third y) 'rrequest) (list '=rrequest> (second y) '=value))))
      (eval `(add-dm (,test isa chunk)))
    (eval
`(p ,(new-name-fct (format nil "action-~A->~A-" (first x)(first y)))
    =goal>
      isa goal
      ,@goal-test
      ,@retrieval-test
      ,@visual-test
   =retrieval>
     isa action
     a ,test
     anext =anext
   ,@imaginal-test-1
   ,@imaginal-test-2
   ,@imaginalreq-test
;   ,@motor-test
   ,@external-action-test
   ,@retrieval-match
==>
   ,@goal-action
   ,@retrieval-action
   ,@imaginalreq-action
   ,@imaginal-action
;   ,@motor-action
   ,@external-action
   +retrieval>
    isa action
    name      =anext))

)))))


;;; Obsolete
(p action-done
   =goal>
     isa goal
   =retrieval>
     isa action
     a done
   ?action>
     state free
==>
   -goal>
   -imaginal>
   -perception>
  !safe-eval! (setf (dm-finsts (get-module declarative)) nil) ;;; Set it up for the next cycle   
   !stop!
)

(p action-goal-finish
   =goal>
      isa goal
      task finish
      condition nil
      action nil
   ?retrieval>
     state free
     buffer empty
   ?imaginal>
     state free
   ?action>
     state free
==>
   -goal>
   -imaginal>
   -perception>
  !safe-eval! (setf (dm-finsts (get-module declarative)) nil) ;;; Set it up for the next cycle   
   !stop!
)

(spp action-goal-finish :u 100)

;;; Generate productions that handle all module actions

(dotimes (action 4)  ;;; 0, 1, 2 or 3 action slots
  (dotimes (newwm 5) ;;; 0, 1..4 slots
    (dolist (harvest '(t nil)) ;; has the retrieval be used -> then clear it.
      
    (dolist (id '(t nil)) ;;; the retrieval can be any combination of id and slots, so we generate them all
      (dolist (slot1 '(t nil))
        (dolist (slot2 '(t nil))
          (dolist (slot3 '(t nil))
            (dolist (slot4 '(t nil))
              (let ((rrequest (or id slot1 slot2 slot3 slot4)))
                (when (not (and harvest rrequest)) ;; if there is a harvest-request and a new retrieval we don't need to handle that separately
                (eval
`(p ,(new-name-fct "check-done-start-modules")
   =goal>
     isa goal
     retrieval-request ,rrequest
     action-request ,(> action 0)
     episodial-request ,(> newwm 0)
     ,@(when (not rrequest) (list 'harvest-request harvest))
   =retrieval>
     isa action
     a stop
   ,@(when harvest '(=rharvest> isa fact))
   ,@(when rrequest (list
   '=rrequest>
     'isa 'fact
     'id (if id '=r0 nil)
     'slot1 (if slot1 '=r1 nil)
     'slot2 (if slot2 '=r2 nil)
     'slot3 (if slot3 '=r3 nil)
     'slot4 (if slot4 '=r4 nil)))
   ,@(when (> action 0) (list
     '=action>
       'isa 'fact
       'slot1 '=a1
       'slot2 (if (> action 1) '=a2 nil)
       'slot3 (if (> action 2) '=a3 nil)
     '?action> 'state 'free))
   ,@(when (> newwm 0) (list
      '=imaginalreq>
        'isa 'fact
        'slot1 '=wm1
        'slot2 (if (> newwm 1) '=wm2 nil)
        'slot3 (if (> newwm 2) '=wm3 nil)
        'slot4 (if (> newwm 3) '=wm4 nil)
      '?imaginal> 'state 'free))
 ;  !safe-eval! (not (dm-failed (get-module declarative)))
==>
  =goal>
     retrieval-request nil
     action-request nil
     episodial-request nil
     harvest-request nil
     action nil
  ,@(when rrequest (flatten (list
 '+retrieval>
     'isa 'fact
     (if id '(id =r0) nil)
     (if slot1 '(slot1 =r1) nil)
     (if slot2 '(slot2 =r2) nil)
     (if slot3 '(slot3 =r3) nil)
     (if slot4 '(slot4 =r4) nil)
  ; '-rharvest>
  ))
      )
  ,@(when rrequest '(=rrequest> slot1 nil slot2 nil slot3 nil slot4 nil))
  ,@(when harvest
      '(=rharvest> id nil slot1 nil slot2 nil slot3 nil slot4 nil)
      )  ;; if there was a harvest set the buffer to empty, otherwise leave it alone
  ,@(if (> action 0) (flatten (list
  '+action>
     'isa 'external-action
      'ac '=a1
      (if (> action 1) '(slot1 =a2) nil)
      (if (> action 2) '(slot2 =a3) nil)))
      '(+perception> isa update-perception))
  ,@(when (> newwm 0) '(=imaginalreq> id nil slot1 nil slot2 nil slot3 nil slot4 nil))
   ,@(when (> newwm 0)(flatten (list 
   '+imaginal>
      'isa 'fact
      'slot1 '=wm1
      (if (> newwm 1) '(slot2 =wm2) nil)
      (if (> newwm 2) '(slot3 =wm3) nil)
      (if (> newwm 3) '(slot4 =wm4) nil))))
   !safe-eval! (setf (dm-finsts (get-module declarative)) nil) 

))))))))))))

#|
(p action-stop-done
   "Last action and there is no retrieval, action or anything else left"
   =goal>
     isa goal
     action wait-all-done
     retrieval-request nil
     action-request nil
     episodial-request nil
   ?retrieval>
     state free
     buffer empty
;   =imaginal>
;     isa fact
;     id =value
   !safe-eval! (not (dm-failed (get-module declarative)))
==>
  =goal>
    action nil

  !safe-eval! (setf (dm-finsts (get-module declarative)) nil) 
)
|#
;;; When we are done with all the actions we are going to handle any retrievals
;;; Retrieval requests and harvests can have any combination of variables and nils, so we need a separate rule for each of them
(dolist (id '(t nil))
(dolist (slot1 '(t nil))
  (dolist (slot2 '(t nil))
    (dolist (slot3 '(t nil))
      (dolist (slot4 '(t nil))
        (let ((name (new-name-fct "harvest-retrieval")))
               (eval
`(p ,name
   =goal>
     isa goal
 ;    action wait-all-done
   =retrieval>
     isa fact
     id ,(if id '=r0 nil)
     slot1 ,(if slot1 '=r1 nil)
     slot2 ,(if slot2 '=r2 nil)
     slot3 ,(if slot3 '=r3 nil)
     slot4 ,(if slot4 '=r4 nil)
   =rrequest>
     isa fact
   =rharvest>
     isa fact
;   !eval! t  ; handle this differently: in the production compilation code
==>
  =goal>
     harvest-request nil
  =rharvest>
     id ,(if id '=r0 nil)
     slot1 ,(if slot1 '=r1 nil)
     slot2 ,(if slot2 '=r2 nil)
     slot3 ,(if slot3 '=r3 nil)
     slot4 ,(if slot4 '=r4 nil)
  =rrequest>
     id ,(if id '=r0 nil)
     slot1 ,(if slot1 '=r1 nil)
     slot2 ,(if slot2 '=r2 nil)
     slot3 ,(if slot3 '=r3 nil)
     slot4 ,(if slot4 '=r4 nil)
  +rrequest>
    isa fact
))
          (eval `(spp ,name :at 0.0))) ;;; action time is zero because this is not a "real" production, just needed to get things to workl


)))))


(p retrieval-failure
   =goal>
     isa goal
     action nil
   ?retrieval>
     state error
   =rrequest>
     isa fact
   =rharvest>
     isa fact
==>
  =rharvest>
    id nil
    slot1 error
    slot2 nil
    slot3 nil
    slot4 nil
  =rrequest>
  	id nil
    slot1 error
    slot2 nil
    slot3 nil
    slot4 nil
  +rrequest>
    isa fact
   !eval! (setf (dm-failed (get-module declarative)) nil)
)


;;; Imaginalreq actions
;; Need versions for different numbers of slots




(spp general-condition-does-not-match :u -20.0 )
(spp general-action-does-not-apply :u -20.0)

)))

(format t "~%Loaded Actransfer version ~A from ~A~%" *actransfer-version* *actransfer-date*)
